"""
This is a module containing generic REST tasks.
"""

import json
from enum import Enum
from typing import TYPE_CHECKING, Any, Dict, Union

import httpx
from prefect import task
from pydantic import VERSION as PYDANTIC_VERSION

if PYDANTIC_VERSION.startswith("2."):
    from pydantic.v1 import BaseModel
else:
    from pydantic import BaseModel

if TYPE_CHECKING:
    from prefect_hex import HexCredentials


class HTTPMethod(Enum):
    """
    Available HTTP request methods.
    """

    GET = "get"
    POST = "post"
    PUT = "put"
    DELETE = "delete"
    PATCH = "patch"


def serialize_model(obj: Any) -> Any:
    """
    Recursively serializes `pydantic.BaseModel` into JSON;
    returns original obj if not a `BaseModel`.

    Args:
        obj: Input object to serialize.

    Returns:
        Serialized version of object.
    """
    if isinstance(obj, list):
        return [serialize_model(o) for o in obj]
    elif isinstance(obj, Dict):
        return {k: serialize_model(v) for k, v in obj.items()}

    if isinstance(obj, BaseModel):
        obj = obj.dict()
    return obj


def strip_kwargs(**kwargs: Dict) -> Dict:
    """
    Recursively drops keyword arguments if value is None,
    and serializes any `pydantic.BaseModel` types.

    Args:
        **kwargs: Input keyword arguments.

    Returns:
        Stripped version of kwargs.
    """
    stripped_dict = {}
    for k, v in kwargs.items():
        v = serialize_model(v)
        if isinstance(v, dict):
            v = strip_kwargs(**v)
        if v is not None:
            stripped_dict[k] = v
    return stripped_dict or {}


@task
async def execute_endpoint(
    endpoint: str,
    hex_credentials: "HexCredentials",
    http_method: HTTPMethod = HTTPMethod.GET,
    params: Dict[str, Any] = None,
    json: Dict[str, Any] = None,
    **kwargs: Dict[str, Any],
) -> httpx.Response:
    """
    Generic function for executing REST endpoints.

    Args:
        endpoint: The endpoint route.
        hex_credentials: Credentials to use for authentication with Hex.
        http_method: Either GET, POST, PUT, DELETE, or PATCH.
        params: URL query parameters in the request.
        json: JSON serializable object to include in the body of the request.
        **kwargs: Additional keyword arguments to pass.

    Returns:
        The httpx.Response from interacting with the endpoint.

    Examples:
        Queries project runs for a given project ID.
        ```python
        from prefect import flow
        from prefect_hex import HexCredentials
        from prefect_hex.rest import execute_endpoint

        @flow
        def example_execute_endpoint_flow():
            endpoint = f"/project/5a8591dd-4039-49df-9202-96385ba3eff8/runs"
            hex_credentials = HexCredentials(token="a1b2c3d4")
            params = dict(limit=100)

            response = execute_endpoint(endpoint, hex_credentials, params=params)
            return response.json()

        example_execute_endpoint_flow()
        ```
    """
    if isinstance(http_method, HTTPMethod):
        http_method = http_method.value

    if params is not None:
        stripped_params = strip_kwargs(**params)
    else:
        stripped_params = None

    if json is not None:
        kwargs["json"] = strip_kwargs(**json)

    async with hex_credentials.get_client() as client:
        response = await getattr(client, http_method)(
            endpoint, params=stripped_params, **kwargs
        )

    return response


def _unpack_contents(response: httpx.Response) -> Union[Dict[str, Any], bytes]:
    """
    Helper method to unpack the contents from the httpx.Response,
    reporting errors in a helpful manner, if any.
    """
    try:
        response.raise_for_status()
    except httpx.HTTPStatusError as exc:
        raise httpx.HTTPStatusError(
            response.json(), request=exc.request, response=exc.response
        ) from exc

    try:
        return response.json()
    except json.JSONDecodeError:
        return response.content



"""Credential classes used to perform authenticated interactions with Hex"""

from httpx import AsyncClient
from prefect.blocks.core import Block
from pydantic import VERSION as PYDANTIC_VERSION

if PYDANTIC_VERSION.startswith("2."):
    from pydantic.v1 import Field, SecretStr
else:
    from pydantic import Field, SecretStr


class HexCredentials(Block):
    """
    Block used to manage Hex authentication.

    Attributes:
        domain: Domain to make API requests against.
        token: The token to authenticate with Hex.


    Examples:
        Load stored Hex credentials:
        ```python
        from prefect_hex import HexCredentials

        hex_credentials_block = HexCredentials.load("BLOCK_NAME")
        ```
    """

    _block_type_name = "Hex Credentials"
    _logo_url = "https://images.ctfassets.net/gm98wzqotmnx/3biMverMLGiDA7y5fkqKZF/4b7747052b59fa8182a9686b88ea9541/Hex_Purple__for_light_backgrounds_.png?h=250"  # noqa
    _documentation_url = "https://prefecthq.github.io/prefect-hex/credentials/#prefect_hex.credentials.HexCredentials"  # noqa

    domain: str = Field(
        default="app.hex.tech", description="Domain to make API requests against."
    )
    token: SecretStr = Field(default=..., description="Token used for authentication.")

    def get_client(self) -> AsyncClient:
        """
        Gets a Hex REST AsyncClient.

        Returns:
            A Hex REST AsyncClient.

        Example:
            Gets a Hex REST AsyncClient.
            ```python
            from prefect import flow
            from prefect_hex import HexCredentials

            @flow
            def example_get_client_flow():
                token = "consumer_key"
                hex_credentials = HexCredentials(token=token)
                client = hex_credentials.get_client()
                return client

            example_get_client_flow()
            ```
        """
        client_kwargs = {
            "base_url": f"https://{self.domain}/api/v1",
            "headers": {"Authorization": f"Bearer {self.token.get_secret_value()}"},
        }
        client = AsyncClient(**client_kwargs)
        return client



"""
This is a module containing exceptions used within prefect-hex.
"""

from prefect_hex.models.project import ProjectRunStatus


class HexProjectRunError(RuntimeError):
    """
    A generic Hex project exception.
    """


class HexProjectRunTimedOut(HexProjectRunError):
    """
    Raised when Hex project run does not complete in the configured max
    wait seconds.
    """


class HexProjectRunUnableToAllocateKernel(HexProjectRunError):
    """
    Raised when Hex project run is unable to allocate a kernel.
    """


class HexProjectRunErrored(HexProjectRunError):
    """
    Raised when Hex project run has errored out.
    """


class HexProjectRunKilled(HexProjectRunError):
    """
    Raised when Hex project run is killed.
    """


TERMINAL_STATUS_EXCEPTIONS = {
    ProjectRunStatus.unabletoallocatekernel: HexProjectRunUnableToAllocateKernel,
    ProjectRunStatus.errored: HexProjectRunErrored,
    ProjectRunStatus.killed: HexProjectRunKilled,
    ProjectRunStatus.completed: None,
}



from httpx import AsyncClient

from prefect_hex import HexCredentials


def test_hex_credentials_get_client():
    client = HexCredentials(domain="domain", token="token_value").get_client()
    assert isinstance(client, AsyncClient)
    assert client.headers["authorization"] == "Bearer token_value"



from typing import List

import httpx
import pytest
from pydantic import VERSION as PYDANTIC_VERSION

if PYDANTIC_VERSION.startswith("2."):
    from pydantic.v1 import BaseModel, Extra
else:
    from pydantic import BaseModel, Extra

from prefect_hex import HexCredentials
from prefect_hex.rest import HTTPMethod, execute_endpoint, serialize_model, strip_kwargs


@pytest.mark.parametrize("params", [dict(a="A", b="B"), None])
@pytest.mark.parametrize("http_method", ["get", HTTPMethod.GET, "post"])
async def test_execute_endpoint(params, http_method, respx_mock):
    url = "https://prefect.io/"

    respx_mock.get(url).mock(return_value=httpx.Response(200))
    respx_mock.post(url).mock(return_value=httpx.Response(200))

    execute_kwargs = dict()
    if http_method == "post":
        execute_kwargs["json"] = {"key": "val"}

    credentials = HexCredentials(
        domain="domain",
        token="token_value",
    )
    response = await execute_endpoint.fn(
        url, credentials, http_method=http_method, params=params, **execute_kwargs
    )
    assert response.status_code == 200


def test_strip_kwargs():
    assert strip_kwargs(**{"a": None, "b": None}) == {}
    assert strip_kwargs(**{"a": "", "b": None}) == {"a": ""}
    assert strip_kwargs(**{"a": "abc", "b": "def"}) == {"a": "abc", "b": "def"}
    assert strip_kwargs(a="abc", b="def") == {"a": "abc", "b": "def"}
    assert strip_kwargs(**dict(a=[])) == {"a": []}


class TestAnotherBaseModel(BaseModel):
    some_float: float
    some_bool: bool


class TestBaseModel(BaseModel):
    class Config:
        extra = Extra.allow
        allow_mutation = False

    some_string: str
    some_int: int
    another_base_model: TestAnotherBaseModel
    other_base_models: List[TestAnotherBaseModel]


def test_serialize_model():
    expected = {
        "base_model": {
            "some_string": "abc",
            "some_int": 1,
            "another_base_model": {"some_float": 2.8, "some_bool": True},
            "other_base_models": [
                {"some_float": 8.8, "some_bool": False},
                {"some_float": 1.8, "some_bool": True},
            ],
            "unexpected_value": ["super", "unexpected"],
        }
    }

    actual = serialize_model(
        {
            "base_model": TestBaseModel(
                some_string="abc",
                some_int=1,
                unexpected_value=["super", "unexpected"],
                another_base_model=TestAnotherBaseModel(some_float=2.8, some_bool=True),
                other_base_models=[
                    TestAnotherBaseModel(some_float=8.8, some_bool=False),
                    TestAnotherBaseModel(some_float=1.8, some_bool=True),
                ],
            )
        }
    )

    assert expected == actual



"""
This is a module containing tasks for interacting with
Hex projects
"""

import asyncio
from typing import Dict, Optional, Tuple

from prefect import flow, get_run_logger, task

from prefect_hex import HexCredentials
from prefect_hex.exceptions import (
    TERMINAL_STATUS_EXCEPTIONS,
    HexProjectRunError,
    HexProjectRunTimedOut,
)
from prefect_hex.models import project as models
from prefect_hex.rest import HTTPMethod, _unpack_contents, execute_endpoint


@task
async def run_project(
    project_id: str,
    hex_credentials: HexCredentials,
    input_params: Optional[Dict] = None,
    dry_run: bool = False,
    update_cache: bool = False,
) -> models.ProjectRunResponsePayload:  # pragma: no cover
    """
    Trigger a run of the latest published version of a project.

    Args:
        project_id:
            Project ID to run.
        hex_credentials:
            Credentials to use for authentication with Hex.
        input_params:
            Optional input parameter value map for this project run, e.g.
            ```
            {"text_input_1": "Hello World", "numeric_input_1": 123}
            ```
        dry_run:
            If specified, perform a dry run without actually executing the project.
        update_cache:
            When true, this run will update the cached state of the published app
            with the latest run results. Additionally, any SQL cells
            that have caching enabled will be re-executed as part of
            this run. Note that this cannot be set to true if custom
            input parameters are provided.

    Returns:
        Information about the triggered project run.
    """  # noqa
    endpoint = f"/project/{project_id}/run"  # noqa

    response = await execute_endpoint.fn(
        endpoint,
        hex_credentials,
        http_method=HTTPMethod.POST,
        json=models.RunProjectRequestBody(
            dryRun=dry_run, inputParams=input_params, updateCache=update_cache
        ).dict(by_alias=True),
    )

    contents = _unpack_contents(response)
    return models.ProjectRunResponsePayload.parse_obj(contents)


@task
async def get_run_status(
    project_id: str,
    run_id: str,
    hex_credentials: HexCredentials,
) -> models.ProjectStatusResponsePayload:  # pragma: no cover
    """
    Get the status of a project run.

    Args:
        project_id:
            Project ID associated with the run to get the status of.
        run_id:
            Run ID of the run to get the status of.
        hex_credentials:
            Credentials to use for authentication with Hex.

    Returns:
        Information about the requested run.
    """  # noqa
    endpoint = f"/project/{project_id}/run/{run_id}"  # noqa

    response = await execute_endpoint.fn(
        endpoint,
        hex_credentials,
        http_method=HTTPMethod.GET,
    )

    contents = _unpack_contents(response)
    return models.ProjectStatusResponsePayload.parse_obj(contents)


@task
async def cancel_run(
    project_id: str,
    run_id: str,
    hex_credentials: HexCredentials,
) -> None:  # pragma: no cover
    """
    Cancel a project run.

    Args:
        project_id:
            Project ID associated with the run to cancel.
        run_id:
            Run ID of the run to cancel.
        hex_credentials:
            Credentials to use for authentication with Hex.
    """  # noqa
    endpoint = f"/project/{project_id}/run/{run_id}"  # noqa

    response = await execute_endpoint.fn(
        endpoint,
        hex_credentials,
        http_method=HTTPMethod.DELETE,
    )

    # Handles any errors returned by the API
    _unpack_contents(response)


@task
async def get_project_runs(
    project_id: str,
    hex_credentials: HexCredentials,
    limit: Optional[models.PageSize] = None,
    offset: Optional[models.Offset] = None,
    status_filter: Optional[models.ProjectRunStatus] = None,
) -> models.ProjectRunsResponsePayload:  # pragma: no cover
    """
    Get the status of the API-triggered runs of a project.

    Args:
        project_id:
            Project ID to get runs for.
        hex_credentials:
            Credentials to use for authentication with Hex.
        limit:
            Number of results to fetch per page for paginated requests.
        offset:
            Offset for paginated requests.
        status_filter:
            Current status of a project run.

    Returns:
        Details of all the retrieved runs.
    """  # noqa
    endpoint = f"/project/{project_id}/runs"  # noqa

    params = {
        "limit": limit,
        "offset": offset,
        "statusFilter": status_filter.value if status_filter is not None else None,
    }

    response = await execute_endpoint.fn(
        endpoint,
        hex_credentials,
        http_method=HTTPMethod.GET,
        params=params,
    )

    contents = _unpack_contents(response)
    return models.ProjectRunsResponsePayload.parse_obj(contents)


@flow
async def trigger_project_run_and_wait_for_completion(
    project_id: str,
    hex_credentials: HexCredentials,
    input_params: Optional[Dict] = None,
    update_cache: bool = False,
    max_wait_seconds: int = 900,
    poll_frequency_seconds: int = 10,
) -> models.ProjectRunResponsePayload:
    """
    Flow that triggers a project run and waits for the triggered run to complete.

    Args:
        project_id:
            Project ID to run.
        hex_credentials:
            Credentials to use for authentication with Hex.
        input_params:
            Optional input parameter value map for this project run, e.g.
            ```
            {"text_input_1": "Hello World", "numeric_input_1": 123}
            ```
        update_cache:
            When true, this run will update the cached state of the published app
            with the latest run results. Additionally, any SQL cells
            that have caching enabled will be re-executed as part of
            this run. Note that this cannot be set to true if custom
            input parameters are provided.
        max_wait_seconds: Maximum number of seconds to wait for the entire
            flow to complete.
        poll_frequency_seconds: Number of seconds to wait in between checks for
            run completion.

    Returns:
        Information about the triggered project run.

    Examples:
        Trigger a Hex project run and wait for completion as a stand-alone flow.
        ```python
        import asyncio
        from prefect_hex import HexCredentials
        from prefect_hex.project import trigger_project_run_and_wait_for_completion

        asyncio.run(
            trigger_sync_run_and_wait_for_completion(
                hex_credentials=HexCredentials(
                    token="1abc0d23-1234-1a2b-abc3-12ab456c7d8e"
                ),
                project_id="012345c6-b67c-1234-1b2c-66e4ad07b9f3",
                max_wait_seconds=1800,
                poll_frequency_seconds=5,
            )
        )
        ```

        Trigger a Hex project run and wait for completion as a subflow.
        ```python
        from prefect import flow
        from prefect_hex import HexCredentials
        from prefect_hex.project import trigger_project_run_and_wait_for_completion

        @flow
        def trigger_project_run_and_wait_for_completion_flow(project_id: str):
            hex_credentials = HexCredentials.load("hex-token")
            project_metadata = trigger_project_run_and_wait_for_completion(
                project_id=project_id,
                hex_credentials=hex_credentials
            )
            return project_metadata

        trigger_project_run_and_wait_for_completion_flow(
            project_id="012345c6-b67c-1234-1b2c-66e4ad07b9f3"
        )
        ```
    """
    logger = get_run_logger()

    project_run_future = await run_project.submit(
        project_id=project_id,
        hex_credentials=hex_credentials,
        input_params=input_params,
        update_cache=update_cache,
    )
    project_run = await project_run_future.result()
    run_id = project_run.run_id

    logger.info(
        "Started project %s run %s; visit %s to view the run.",
        repr(project_id),
        repr(run_id),
        str(project_run.run_status_url),
    )

    project_status, project_metadata = await wait_for_project_run_completion(
        project_id=project_id,
        run_id=run_id,
        hex_credentials=hex_credentials,
        max_wait_seconds=max_wait_seconds,
        poll_frequency_seconds=poll_frequency_seconds,
    )

    if project_status == models.ProjectRunStatus.completed:
        return project_metadata
    else:
        raise TERMINAL_STATUS_EXCEPTIONS.get(project_status, HexProjectRunError)(
            f"Project {project_id!r} run {run_id!r} "
            f"was unsuccessful with {project_status.value!r} status"
        )


@flow
async def wait_for_project_run_completion(
    project_id: str,
    run_id: str,
    hex_credentials: HexCredentials,
    max_wait_seconds: int = 900,
    poll_frequency_seconds: int = 10,
) -> Tuple[models.ProjectRunStatus, models.ProjectStatusResponsePayload]:
    """
    Flow that waits for the triggered project run to complete.

    Args:
        project_id:
            Project ID to watch.
        run_id:
            Run ID to wait for.
        hex_credentials:
            Credentials to use for authentication with Hex.
        max_wait_seconds: Maximum number of seconds to wait for the entire
            flow to complete.
        poll_frequency_seconds: Number of seconds to wait in between checks for
            run completion.

    Returns:
        The status of the project run and the metadata associated with the run.

    Examples:
        Wait for completion of a project run as a subflow.
        ```python
        from prefect import flow
        from prefect_hex import HexCredentials
        from prefect_hex.project import wait_for_project_run_completion

        @flow
        def wait_for_project_run_completion_flow(project_id: str, run_id: str):
            hex_credentials = HexCredentials.load("hex-token")
            project_status, project_metadata = wait_for_project_run_completion(
                project_id=project_id,
                run_id=run_id,
                hex_credentials=hex_credentials
            )
            return project_status, project_metadata

        wait_for_project_run_completion_flow(
            project_id="012345c6-b67c-1234-1b2c-66e4ad07b9f3",
            run_id="654321c6-b67c-1234-1b2c-66e4ad07b9f3",
        )
        ```
    """
    logger = get_run_logger()
    seconds_waited_for_run_completion = 0
    wait_for = []

    while seconds_waited_for_run_completion <= max_wait_seconds:
        project_future = await get_run_status.submit(
            project_id=project_id,
            run_id=run_id,
            hex_credentials=hex_credentials,
            wait_for=wait_for,
        )
        wait_for = [project_future]

        project_metadata = await project_future.result()
        project_status = project_metadata.status
        if project_status in TERMINAL_STATUS_EXCEPTIONS.keys():
            return project_status, project_metadata

        logger.debug(
            "Waiting on project %s run %s with sync status %s for %s seconds",
            repr(project_id),
            repr(run_id),
            repr(project_status.value),
            poll_frequency_seconds,
        )
        await asyncio.sleep(poll_frequency_seconds)
        seconds_waited_for_run_completion += poll_frequency_seconds

    raise HexProjectRunTimedOut(
        f"Max wait time of {max_wait_seconds} seconds exceeded while waiting "
        f"for project {project_id!r} run {run_id!r}"
    )



import pytest
from httpx import Response

from prefect_hex.exceptions import TERMINAL_STATUS_EXCEPTIONS, HexProjectRunTimedOut
from prefect_hex.models.project import ProjectStatusResponsePayload
from prefect_hex.project import trigger_project_run_and_wait_for_completion


@pytest.fixture()
def project_run_json():
    return {
        "projectId": "123",
        "runId": "1234",
        "runUrl": "https://app.hex.tech/12345/app/123",
        "runStatusUrl": "https://app.hex.tech/api/v1/project/12345/run/1234",
        "traceId": "123456",
    }


@pytest.fixture()
def project_status_json():
    return {
        "projectId": "123",
        "runId": "1234",
        "status": "COMPLETED",
        "runUrl": "https://app.hex.tech/12345/app/123",
        "startTime": "2022-11-15T23:53:31.554Z",
        "endTime": None,
        "elapsedTime": 1234,
        "traceId": "123456",
    }


async def test_trigger_project_run_and_wait_for_completion(
    hex_credentials, respx_mock, project_run_json, project_status_json
):
    respx_mock.post("https://app.hex.tech/api/v1/project/123/run").mock(
        return_value=Response(200, json=project_run_json)
    )
    respx_mock.get("https://app.hex.tech/api/v1/project/123/run/1234").mock(
        return_value=Response(200, json=project_status_json)
    )
    actual = await trigger_project_run_and_wait_for_completion(
        project_id="123", hex_credentials=hex_credentials
    )
    assert isinstance(actual, ProjectStatusResponsePayload)
    assert actual.project_id == "123"
    assert actual.run_id == "1234"


@pytest.mark.parametrize("status", ["RUNNING", "PENDING"])
async def test_trigger_sync_run_and_wait_for_completion_timeout(
    hex_credentials, respx_mock, project_run_json, project_status_json, status
):
    project_status_json["status"] = status
    respx_mock.post("https://app.hex.tech/api/v1/project/123/run").mock(
        return_value=Response(200, json=project_run_json)
    )
    respx_mock.get("https://app.hex.tech/api/v1/project/123/run/1234").mock(
        return_value=Response(200, json=project_status_json)
    )
    with pytest.raises(HexProjectRunTimedOut, match="Max wait time of 2 seconds"):
        await trigger_project_run_and_wait_for_completion(
            project_id="123",
            hex_credentials=hex_credentials,
            max_wait_seconds=2,
            poll_frequency_seconds=1,
        )


@pytest.mark.parametrize("status", TERMINAL_STATUS_EXCEPTIONS.keys())
async def test_trigger_project_run_and_wait_for_completion_unsuccessful(
    hex_credentials, respx_mock, project_run_json, project_status_json, status
):
    if status.value == "COMPLETED":
        pytest.skip()

    project_status_json["status"] = status.value
    respx_mock.post("https://app.hex.tech/api/v1/project/123/run").mock(
        return_value=Response(200, json=project_run_json)
    )
    respx_mock.get("https://app.hex.tech/api/v1/project/123/run/1234").mock(
        return_value=Response(200, json=project_status_json)
    )
    with pytest.raises(
        TERMINAL_STATUS_EXCEPTIONS[status],
        match=f"was unsuccessful with {status.value!r}",
    ):
        await trigger_project_run_and_wait_for_completion(
            project_id="123",
            hex_credentials=hex_credentials,
        )



from __future__ import annotations

from datetime import datetime
from enum import Enum
from typing import Any, Dict, List, Optional

from pydantic import VERSION as PYDANTIC_VERSION

if PYDANTIC_VERSION.startswith("2."):
    from pydantic.v1 import BaseModel, Extra, Field, HttpUrl
else:
    from pydantic import BaseModel, Extra, Field, HttpUrl


class ProjectRunStatus(Enum):
    """
    Current status of a project run.
    """

    pending = "PENDING"
    running = "RUNNING"
    errored = "ERRORED"
    completed = "COMPLETED"
    killed = "KILLED"
    unabletoallocatekernel = "UNABLE_TO_ALLOCATE_KERNEL"


class Offset(BaseModel):
    class Config:
        extra = Extra.allow
        allow_mutation = False

    __root__: int = Field(
        default=..., description="Offset for paginated requests.", ge=0
    )


class PageSize(BaseModel):
    class Config:
        extra = Extra.allow
        allow_mutation = False

    __root__: int = Field(
        default=...,
        description="Number of results to fetch per page for paginated requests.",
        ge=1,
        le=100,
    )


class RunProjectRequestBody(BaseModel):
    class Config:
        extra = Extra.allow
        extra = Extra.forbid
        allow_mutation = False

    dry_run: Optional[bool] = Field(
        "false",
        alias="dryRun",
        description=(
            "If specified, perform a dry run without actually executing the project."
        ),
    )
    input_params: Optional[Dict[str, Any]] = Field(
        None,
        alias="inputParams",
        description="Optional input parameter value map for this project run.",
        example={"numeric_input_1": 123, "text_input_1": "Hello World"},
    )
    update_cache: Optional[bool] = Field(
        "false",
        alias="updateCache",
        description=(
            "When true, this run will update the cached state of the published app with"
            " the latest run results.\nAdditionally, any SQL cells that have caching"
            " enabled will be re-executed as part of this run. Note\nthat this cannot"
            " be set to true if custom input parameters are provided."
        ),
    )


class ProjectRunResponsePayload(BaseModel):
    """
    Response format returned by the runProject endpoint.
    """

    class Config:
        extra = Extra.allow
        allow_mutation = False

    project_id: str = Field(
        default=..., alias="projectId", description="Unique ID for a Hex project."
    )
    run_id: str = Field(
        default=..., alias="runId", description="Unique ID for a run of a Hex project."
    )
    run_status_url: HttpUrl = Field(
        default=...,
        alias="runStatusUrl",
        description="URL to query the status of the project run via the Hex API.",
    )
    run_url: HttpUrl = Field(
        default=...,
        alias="runUrl",
        description="URL to view the current progress of the project run in the Hex UI.",
    )
    trace_id: str = Field(
        default=...,
        alias="traceId",
        description=(
            "Hex trace ID to identify an API request. Provide this value to hex support"
            " with any API issues you encounter."
        ),
    )


class ProjectStatusResponsePayload(BaseModel):
    """
    Response format returned by the getRunStatus endpoint
    """

    class Config:
        extra = Extra.allow
        allow_mutation = False

    elapsed_time: float = Field(
        default=...,
        alias="elapsedTime",
        description="Total elapsed time for the project run in milliseconds.",
    )
    end_time: Optional[datetime] = Field(
        default=None,
        alias="endTime",
        description="UTC timestamp of when the project run finished.",
    )
    project_id: str = Field(
        default=..., alias="projectId", description="Unique ID for a Hex project."
    )
    run_id: str = Field(
        default=..., alias="runId", description="Unique ID for a run of a Hex project."
    )
    run_url: HttpUrl = Field(
        default=...,
        alias="runUrl",
        description="URL to view the current progress of the project run in the Hex UI.",
    )
    start_time: datetime = Field(
        default=...,
        alias="startTime",
        description="UTC timestamp of when the project run started.",
    )
    status: ProjectRunStatus
    trace_id: str = Field(
        default=...,
        alias="traceId",
        description=(
            "Hex trace ID to identify an API request. Provide this value to hex support"
            " with any API issues you encounter."
        ),
    )


class ProjectRunsResponsePayload(BaseModel):
    """
    Response format returned by the getProjectRuns endpoint.
    """

    class Config:
        extra = Extra.allow
        allow_mutation = False

    next_page: Optional[HttpUrl] = Field(
        None,
        alias="nextPage",
        description="URL to fetch the next page of results for a paginated API request.",
    )
    previous_page: Optional[HttpUrl] = Field(
        None,
        alias="previousPage",
        description="URL to fetch the previous page of results for a paginated API request.",
    )
    runs: List[ProjectStatusResponsePayload] = Field(
        default=...,
        description=(
            "Array of run status payloads in the same format returned by the"
            " `GetRunStatus` endpoint."
        ),
    )
    trace_id: str = Field(
        default=...,
        alias="traceId",
        description=(
            "Hex trace ID to identify an API request. Provide this value to hex support"
            " with any API issues you encounter."
        ),
    )
